mod random;

use crate::state::GameState;
use crate::tree::arena::Arena;
use crate::tree::node::{Node, RewardVal};
use std::collections::HashMap;

/// The simulation policy used in the MCTS simulation phase
///
/// This policy drives the game simulations while evaluating the tree. While
/// a random policy works well, a game-specific policy can be provided either
/// as a custom policy.
#[derive(Debug)]
pub enum SimulationPolicy<S: GameState> {
    /// Random simulation policy
    ///
    /// The sequential actions are selected randomly from the available actions
    /// at each state until a terminal state is found.
    Random,

    /// Custom simulation policy
    Custom(Box<dyn CustomSimulationPolicy<S>>),
}

/// Trait for an object implementing the simulation logic whene exploring the MCTS
/// search tree.
pub trait CustomSimulationPolicy<S: GameState>: std::fmt::Debug {
    /// Simulates the gameplay from the current node onward, returning the rewards
    ///
    /// This should simulate the game until a terminal node is reached, returning
    /// the final reward for each player at the terminal node
    fn simulate(&self, node: &Node<S>, arena: &Arena<S>) -> HashMap<S::Player, RewardVal>;
}

pub fn simulate_reward<S: GameState>(
    policy: &SimulationPolicy<S>,
    node: &Node<S>,
    arena: &Arena<S>,
) -> HashMap<S::Player, RewardVal> {
    match policy {
        SimulationPolicy::Random => random::simulate(node),
        SimulationPolicy::Custom(custom_policy) => custom_policy.simulate(node, arena),
    }
}
