use crate::state::GameState;
use crate::tree::node::Node;

/// An arena for Node allocation
///
/// We use an arena for node allocation to improve performance of our search.
/// The memory is contiguous which allows for faster movement through the tree,
/// as well as more efficient destruction as our MCTS search will destroy the
/// entire tree at once.
pub struct Arena<S: GameState> {
    pub nodes: Vec<Node<S>>,
}

impl<S: GameState> Arena<S> {
    /// Create a new Arena with the given initial capacity
    ///
    /// The arena creates a contiguous block. By reserving an initial capacity
    /// that is sufficient to encapsulate a full search tree we can reduce the
    /// number of reallocs that are required. This number is highly game
    /// dependent.
    pub fn new(initial_capacity: usize) -> Self {
        Arena {
            nodes: Vec::with_capacity(initial_capacity),
        }
    }

    /// Adds a node to the Arena, returning its identifier
    ///
    /// This appends the node to the allocated Arena, and returns the nodes
    /// index in the arena which is used as an identifier for later retrieval.
    pub fn add_node(&mut self, node: Node<S>) -> usize {
        let id = self.nodes.len();
        self.nodes.push(node);
        id
    }

    /// Retrieves a mutable reference to a Node in the Arena
    pub fn get_node_mut(&mut self, id: usize) -> &mut Node<S> {
        &mut self.nodes[id]
    }

    /// Retrieves a reference to a Node in the Arena
    pub fn get_node(&self, id: usize) -> &Node<S> {
        &self.nodes[id]
    }
}
