//! Upper Confidence Bound 1 Tuned (UCB1-Tuned) selection policy
//!
//! This is a fine-tuned version of UCB which takes into account the
//! empircally measured variance of the rewards to drive the exploration.
//!
//! This has been found to perform substantially better than UCB1 in most
//! situations.
//!
//! Auer, P., Cesa-Bianchi, N. & Fischer, P.
//!  Finite-time Analysis of the Multiarmed Bandit Problem.
//!  Machine Learning 47, 235–256 (2002). https://doi.org/10.1023/A:1013689704352

use crate::state::GameState;
use crate::tree::arena::Arena;
use crate::tree::node::{Node, RewardVal};

/// Selects the index of the "best" child using the UCB1-Tuned selection policy
pub fn select_best_child<S: GameState>(
    exploration_constant: f64,
    node: &Node<S>,
    arena: &Arena<S>,
) -> usize {
    if node.is_leaf() {
        panic!("select_best_child called on leaf node");
    }

    let player = node.state.get_current_player();
    let parent_visits = node.visits;
    let best_child = node
        .children
        .iter()
        .max_by(|&a, &b| {
            let node_a = arena.get_node(*a);
            let node_b = arena.get_node(*b);
            let ucb_a = ucb1_tuned_value(
                exploration_constant,
                parent_visits,
                node_a.visits,
                node_a.reward_average(player),
                node_a.reward_variance(player),
            );
            let ucb_b = ucb1_tuned_value(
                exploration_constant,
                parent_visits,
                node_b.visits,
                node_b.reward_average(player),
                node_b.reward_variance(player),
            );
            ucb_a.partial_cmp(&ucb_b).unwrap()
        })
        .unwrap();
    *best_child
}

/// Calculates the UCB1-Tuned value for a node
pub fn ucb1_tuned_value(
    exploration_constant: f64,
    parent_visits: u64,
    child_visits: u64,
    reward_avg: RewardVal,
    reward_variance: f64,
) -> RewardVal {
    if child_visits == 0 {
        RewardVal::INFINITY // Always explore nodes that have never been visited
    } else {
        let parent_visits: RewardVal = parent_visits as RewardVal;
        let child_visits: RewardVal = child_visits as RewardVal;

        // N: number of visits to the parent node
        // n: number of visits to the child node
        // x_i: reward of the ith visit to the child node
        // X: average reward of the child
        // C: exploration constant
        //
        // UCB1-Tuned = X + C * sqrt(Ln(parent_visits) / child_visits * min(1/4, V_n)
        //   V(n) = sum(x_i^2)/n - X^2 + sqrt(2*ln(N)/n)
        let exploitation = reward_avg;
        let mut variance = reward_variance + (2.0 * parent_visits.ln() / child_visits).sqrt();
        if variance > 0.25 {
            variance = 0.25;
        }
        let exploration =
            exploration_constant * (parent_visits.ln() / child_visits * variance).sqrt();

        exploitation + exploration
    }
}
